from html2text import HTML2Text
import bleach
from bs4 import BeautifulSoup


def _get_html2text_obj_with_config() -> HTML2Text:
    """Get HTML2Text object with config set."""
    txt_maker = HTML2Text()

    # some settings to avoid newline problems with links
    txt_maker.ignore_links = False
    txt_maker.body_width = 0
    txt_maker.protect_links = True
    txt_maker.wrap_links = False

    # table settings:
    txt_maker.ignore_tables = True  # keep row content
    return txt_maker


def _get_plaintext_from_html(html: str) -> str:
    """html -> ASCII plaintext, via HTML2Text."""
    txt_maker = _get_html2text_obj_with_config()
    doc = txt_maker.handle(html)
    return doc


def _remove_code(html: str) -> str:
    # exclude 'code' tags from link output:
    soup = BeautifulSoup(html, 'lxml')
    for s in soup.select('code'):
        s.extract()
    html_str = str(soup)
    return html_str


def _remove_del_text(html: str) -> str:
    soup = BeautifulSoup(html, 'lxml')
    for s in soup.select('del'):
        s.extract()
    html_str = str(soup)
    return html_str


def _remove_main_formatting(
        html: str, *,
        tags: list[str] = ['p', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6']) -> str:
    return bleach.clean(html, tags=tags, strip=True)


def _remove_latex(html: str) -> str:
    soup = BeautifulSoup(html, 'lxml')
    for s in soup.select('span', {'class': 'MathJax_Preview'}):
        s.extract()
    html_str = str(soup)
    return html_str


def _get_all_latex_from_html_content(html: str) -> list[str]:
    soup = BeautifulSoup(html, 'html.parser')

    s_content = soup.find_all('span', {'class': 'MathJax_Preview'},
                              text=True)
    latex_found_list = [i.text for i in s_content]
    return latex_found_list



import re
import yaml
from pathlib import Path
from glob import glob
import markdown
import frontmatter
from ._constants import (WIKILINK_REGEX,
                         TAG_MAIN_ONLY_REGEX, TAG_INCLUDE_NESTED_REGEX,
                         WIKILINK_AS_STRING_REGEX,
                         EMBEDDED_FILE_LINK_AS_STRING_REGEX,
                         INLINE_LINK_AFTER_HTML_PROC_REGEX,
                         INLINE_LINK_VIA_MD_ONLY_REGEX)
from ._io import (get_relpaths_from_dir,
                  get_relpaths_matching_subdirs)
from .html_processing import (_get_plaintext_from_html,
                              _remove_code, _remove_latex, _remove_del_text,
                              _remove_main_formatting,
                              _get_all_latex_from_html_content)


def get_md_relpaths_from_dir(dir_path: Path) -> list[Path]:
    """Get list of relative paths for markdown files in a given directory,
    including any subdirectories.

    Thus if the vault directory is the argument, then the function returns
    a list of all the md files found in the vault.

    Args:
        dir_path (pathlib Path): Path object representing the directory
            to search.

    Returns:
        list of Path objects
    """
    return get_relpaths_from_dir(dir_path, extension='md')


def get_md_relpaths_matching_subdirs(dir_path: Path, *,
                                     include_subdirs: list = None,
                                     include_root: bool = True) -> list[Path]:
    """Get list of relative paths for markdown files in a given directory,
    filtered to include specified subdirectories (with include_subdirs
    kwarg).  The default arguments align with get_md_relpaths_from_dir
    function, but this function enables more flexibility.

    For example, if you had a vault with folders named by category, and
    filter them like this in Obsidian:
        path:Category1/ OR path:Category2/ OR path:Category4/
    then you can use the include_subdirs kwarg to do that with this function:
        include_subdirs = ['Category1', 'Category2', 'Category4']

    You can also specify deeper levels to filter on, e.g.:
        include_subdirs = ['Category1/TopicA', 'Category1/TopicB']

    Args:
        dir_path (pathlib Path): Path object representing the directory
            to search.
        include_subdirs (list, optional): list of string paths to include
            in the filtered list of md files (e.g. ['p1', 'p2', 'p3/sp1']).
            If no list is specified, then no filtering is done on paths.
            Defaults to None.
        include_root (bool, optional): include files that are directly in
            the dir_path (root dir).  Defaults to True.

    Returns:
        list of Path objects
    """
    return get_relpaths_matching_subdirs(
        dir_path,
        extension='md',
        include_subdirs=include_subdirs,
        include_root=include_root)


def get_wikilinks(filepath: Path) -> list[str]:
    """Get ALL wikilinks from a md file.
    The links' order of appearance in the file IS preserved in the output.

    This accounts for:
    - Aliases / alt text, so [[Lorem ipsum|L.I.]]
    will be represented as 'Lorem ipsum'.
    - Header text links, so [[Lorem ipsum#Dummy text]]
    will be represented as 'Lorem ipsum'.

    The links' order of appearance in the file IS preserved in the output.

    Args:
        filepath (pathlib Path): Path object representing the file from
            which info will be extracted.

    Returns:
        list of strings
    """
    src_txt = get_source_text_from_md_file(filepath, remove_code=True)

    wikilinks = _get_all_wikilinks_from_source_text(
        src_txt, remove_aliases=True)
    return wikilinks


def get_embedded_files(filepath: Path) -> list[str]:
    """Get ALL embedded files from a md file.
    The embedded files' order of appearance in the file IS preserved in the output.

    This accounts for:
    - Aliases / alt text, so [[Lorem ipsum|L.I.]]
    will be represented as 'Lorem ipsum'.

    The links' order of appearance in the file IS preserved in the output.

    Args:
        filepath (pathlib Path): Path object representing the file from
            which info will be extracted.

    Returns:
        list of strings
    """
    src_txt = get_source_text_from_md_file(filepath, remove_code=True)

    files = _get_all_embedded_files_from_source_text(
        src_txt, remove_aliases=True)
    return files


def get_unique_wikilinks(filepath: Path) -> list[str]:
    """Get UNIQUE wikilinks from a md file.
    The links' order of appearance in the file IS preserved in the output.

    This accounts for:
    - Aliases / alt text, so [[Lorem ipsum|L.I.]]
    will be represented as 'Lorem ipsum'.
    - Header text links, so [[Lorem ipsum#Dummy text]]
    will be represented as 'Lorem ipsum'.

    Args:
        filepath (pathlib Path): Path object representing the file from
            which info will be extracted.

    Returns:
        list of strings
    """
    src_txt = get_source_text_from_md_file(filepath, remove_code=True)

    wikilinks = _get_unique_wikilinks_from_source_text(src_txt, remove_aliases=True)
    return wikilinks


def get_md_links(filepath: Path) -> list[str]:
    """Get markdown links from a md file.
    The links' order of appearance in the file IS preserved in the output.

    This is to check for syntax of the format [...](...).
    The returned 'links' inside the () are not checked for validity or
    subtle differences (e.g. '/' vs no '/' at the end of a URL).

    Args:
        filepath (pathlib Path): Path object representing the file from
            which info will be extracted.

    Returns:
        list of strings
    """
    src_txt = get_source_text_from_md_file(filepath, remove_code=True)
    return _get_md_links_from_source_text(src_txt)


def _get_md_links_from_source_text(src_txt: str) -> list[str]:
    links = _get_all_md_link_info_from_source_text(src_txt)
    if links:  # return links only, not their text
        return [t[-1] for t in links]
    else:
        return links  # empty list


def get_unique_md_links(filepath: Path) -> list[str]:
    """Get markdown links (unique) from a md file.
    The links' order of appearance in the file IS preserved in the output.

    This is to check for syntax of the format [...](...).
    The returned 'links' inside the () are not checked for validity or
    subtle differences (e.g. '/' vs no '/' at the end of a URL).

    Args:
        filepath (pathlib Path): Path object representing the file from
            which info will be extracted.

    Returns:
        list of strings
    """
    src_txt = get_source_text_from_md_file(filepath, remove_code=True)

    links = _get_unique_md_links_from_source_text(src_txt)
    return links


def get_front_matter(filepath: Path) -> dict:
    """Get front matter from a md file.

    If no front matter is found for a file, the value will be {}.

    Args:
        filepath (pathlib Path): Path object representing the file from
            which info will be extracted.
    Returns:
        dict
    """
    front_matter, _ = _get_md_front_matter_and_content(filepath)
    return front_matter


def get_tags(filepath: Path, *, show_nested: bool = False) -> list[str]:
    """Get tags from a md file, based on the order they appear in the file.
    By default, only the highest level of any nested tags is shown in the
    output.

    If no tags are found for a file, the value will be [].

    Args:
        filepath (pathlib Path): Path object representing the file from
            which info will be extracted.
        show_nested (Boolean): show nested tags in the output.  Defaults to
            False (which would mean only the highest level of any nested tags
            are included in the output).
    Returns:
        list
    """
    # get text from source file, but remove any '\#' and code:
    src_txt = get_source_text_from_md_file(
        filepath, remove_code=True,
        str_transform_func=_transform_md_file_string_for_tag_parsing)
    # remove wikilinks so that '#' headers are not caught:
    src_txt = _remove_wikilinks_from_source_text(src_txt)
    tags = _get_tags_from_source_text(src_txt, show_nested=show_nested)
    return tags


def _get_md_front_matter_and_content(filepath: Path, *,
                                     str_transform_func=None) -> tuple[dict, str]:
    """parse md file into front matter and note content"""
    with open(filepath, encoding='utf-8') as f:
        try:
            file_string = f.read()
            if str_transform_func:
                file_string = str_transform_func(file_string)
            return frontmatter.parse(file_string)
        # for invalid YAML, return the whole file as content:
        except yaml.scanner.ScannerError as e:
            print(f"Front matter not populated for {filepath.name}: {repr(e)}")
            return {}, file_string
        except yaml.parser.ParserError as e:
            print(f"Front matter not populated for {filepath.name}: {repr(e)}")
            return {}, file_string
        # handle template {{}} chars in front matter:
        except yaml.constructor.ConstructorError:
            file_string_esc = file_string.translate(
                str.maketrans({"{": r"\{",
                               "}": r"\}"}))
            return frontmatter.parse(file_string_esc)
        # any others:
        except:
            return {}, file_string


def _get_html_from_md_file(filepath: Path, *,
                           str_transform_func=None) -> str:
    """md file -> html (without front matter).

    pymarkdown extensions are used and configured to reflect the Obsidian
    experience as much as possible.  For example, arithmatex is necessary
    to parse math, md_mermaid for mermaid diagram support, etc.
    """
    _, md_content = _get_md_front_matter_and_content(
        filepath,
        str_transform_func=str_transform_func)
    html = _get_html_from_md_content(md_content)
    return html


def _get_html_from_md_content(md_content: str) -> str:
    """md content -> html (without front matter)"""
    html = markdown.markdown(md_content, output_format='html',
                             extensions=['pymdownx.arithmatex',
                                         'pymdownx.superfences',
                                         'pymdownx.mark',
                                         'pymdownx.tilde',
                                         'pymdownx.saneheaders',
                                         'footnotes',
                                         'sane_lists',
                                         'tables'],
                             extension_configs={'pymdownx.tilde':
                                                {'subscript': False}})
    return html


def get_source_text_from_html(html: str, *,
                              remove_code: bool = False,
                              remove_math: bool = False) -> str:
    """html (without front matter) -> ASCII plaintext"""
    if remove_code:
        html = _remove_code(html)
    if remove_math:
        html = _remove_latex(html)
    return _get_plaintext_from_html(html)


def get_source_text_from_md_file(filepath: Path, *,
                                 remove_code: bool = False,
                                 remove_math: bool = False,
                                 str_transform_func=None) -> str:
    """md file -> html (without front matter) -> ASCII plaintext"""
    # strip out front matter (if any):
    html = _get_html_from_md_file(
        filepath,
        str_transform_func=str_transform_func)

    return get_source_text_from_html(html, remove_code=remove_code,
                                     remove_math=remove_math)


def get_readable_text_from_md_file(filepath: Path, *,
                                   tags: list[str] = None) -> str:
    """md file -> html -> plaintext with major formatting removed."""
    # strip out front matter (if any):
    html = _get_html_from_md_file(
        filepath)
    # wikilinks and md links as text:
    html = _replace_md_links_with_their_text(html)
    html = _replace_wikilinks_with_their_text(html)
    html = _remove_embedded_file_links_from_text(html)
    # remove code and remove major formatting on text:
    html = _remove_code(html)
    html = _remove_latex(html)
    html = _remove_del_text(html)
    if tags is not None:
        html = _remove_main_formatting(html, tags=tags)
    else:  # defaults
        html = _remove_main_formatting(html)

    return _get_plaintext_from_html(html)


def _get_all_wikilinks_and_embedded_files(src_txt: str) -> list[str]:
    # extract links
    pattern = re.compile(WIKILINK_REGEX)

    link_matches_list = pattern.findall(src_txt)
    return link_matches_list


def _remove_aliases_from_wikilink_regex_matches(link_matches_list: list[str]) -> list[str]:
    return [(i.replace('\\', '')
             .split("|")[0].rstrip()  # catch alias/alt-text
             .split('#', 1)[0])  # catch links to headers
            for i in link_matches_list]


def _get_all_wikilinks_from_source_text(src_txt: str, *,
                                        remove_aliases: bool = True) -> list[str]:
    matches_list = _get_all_wikilinks_and_embedded_files(src_txt)
    link_matches_list = [g[1] for g in matches_list
                         if g[0] == '']

    if remove_aliases:
        link_matches_list = _remove_aliases_from_wikilink_regex_matches(
            link_matches_list)

    # remove .md:
    link_matches_list = [name.removesuffix('.md')
                         for name in link_matches_list]
    return link_matches_list


def _get_all_embedded_files_from_source_text(src_txt: str, *,
                                             remove_aliases: bool = True) -> list[str]:
    matches_list = _get_all_wikilinks_and_embedded_files(src_txt)
    embedded_files_sublist = [g[1] for g in matches_list
                              if g[0] == '!']

    if remove_aliases:
        embedded_files_sublist = _remove_aliases_from_wikilink_regex_matches(
            embedded_files_sublist)
    return embedded_files_sublist


def _get_all_latex_from_md_file(filepath: Path) -> list[str]:
    return _get_all_latex_from_html_content(
        _get_html_from_md_file(filepath))


def _get_unique_wikilinks_from_source_text(src_txt: str, *,
                                           remove_aliases: bool = True) -> list[str]:
    wikilinks = _get_all_wikilinks_from_source_text(
        src_txt, remove_aliases=remove_aliases)
    return list(dict.fromkeys(wikilinks))


def _get_all_md_link_info_from_source_text(src_txt: str) -> list[tuple[str]]:
    links_regex = re.compile(INLINE_LINK_AFTER_HTML_PROC_REGEX)

    links_list_of_tuples = list(links_regex.findall(src_txt))
    return links_list_of_tuples


def _get_unique_md_links_from_source_text(src_txt: str) -> list[str]:
    links_detail = _get_all_md_link_info_from_source_text(
        src_txt)
    links_list = [link for _, link in links_detail]
    return list(dict.fromkeys(links_list))


def _remove_wikilinks_from_source_text(src_txt: str) -> str:
    return re.sub(WIKILINK_REGEX, '', src_txt)


def _transform_md_file_string_for_tag_parsing(txt: str) -> str:
    return txt.replace('\\#', '')


def _get_tags_from_source_text(src_txt: str, *,
                               show_nested: bool = False) -> list[str]:
    if not show_nested:
        pattern = re.compile(TAG_MAIN_ONLY_REGEX)
    else:
        pattern = re.compile(TAG_INCLUDE_NESTED_REGEX)
    tags_list = pattern.findall(src_txt)
    return tags_list


def _replace_wikilinks_with_their_text(src_txt: str) -> str:
    # get list of wikilinks as strings:
    links_list = _get_all_wikilinks_from_source_text(
        src_txt, remove_aliases=False)

    # get links in their text format:
    readable_text_list = [(i.replace('\\', '')
                           # get wikilinks w/o alias, otherwise alias:
                           .split("|")[-1]
                           .strip())
                          for i in links_list]

    # loop over txt content to replace "[[...]]" wikilinks w/ readable text:
    out_str = src_txt
    links_w_brackets_list = ["".join(["[[", i, "]]"]) for i in links_list]
    switch_dict = dict(zip(links_w_brackets_list, readable_text_list))

    for k, v in switch_dict.items():
        out_str = out_str.replace(k, v)
    return out_str


def _replace_md_links_with_their_text(src_txt: str) -> str:
    # get list of wikilinks as strings:
    matched_text_list = re.findall(WIKILINK_AS_STRING_REGEX, src_txt)
    # get the detail from groups:
    links_detail = re.findall(INLINE_LINK_VIA_MD_ONLY_REGEX, src_txt)

    # get links in their text format:
    readable_text_list = [text for text, _ in links_detail]

    # loop over txt content to replace md links w/ readable text:
    out_str = src_txt
    switch_dict = dict(zip(matched_text_list, readable_text_list))

    for k, v in switch_dict.items():
        out_str = out_str.replace(k, v)
    return out_str


def _remove_embedded_file_links_from_text(src_txt: str) -> str:
    # get list of embedded file links as strings:
    links_list = re.findall(EMBEDDED_FILE_LINK_AS_STRING_REGEX, src_txt)
    # add in the ![[...]] chars:
    links_list = ["".join(['![[', i, ']]']) for i in links_list]

    # remove from text:
    out_str = src_txt
    for i in links_list:
        out_str = out_str.replace(i, '')
    return out_str



import json
import networkx as nx
from pathlib import Path
from ._io import (get_relpaths_from_dir,
                  get_relpaths_matching_subdirs)


def get_canvas_relpaths_from_dir(dir_path: Path) -> list[Path]:
    """Get list of relative paths for canvas files in a given directory,
    including any subdirectories.

    Thus if the vault directory is the argument, then the function returns
    a list of all the canvas files found in the vault.

    Args:
        dir_path (pathlib Path): Path object representing the directory
            to search.

    Returns:
        list of Path objects
    """
    return get_relpaths_from_dir(dir_path, extension='canvas')


def get_canvas_relpaths_matching_subdirs(dir_path: Path, *,
                                         include_subdirs: list = None,
                                         include_root: bool = True) -> list[Path]:
    """Get list of relative paths for canvas files in a given directory,
    filtered to include specified subdirectories (with include_subdirs
    kwarg).  The default arguments align with get_canvas_relpaths_from_dir
    function, but this function enables more flexibility.

    For example, if you had a vault with folders named by category, and
    filter them like this in Obsidian:
        path:Category1/ OR path:Category2/ OR path:Category4/
    then you can use the include_subdirs kwarg to do that with this function:
        include_subdirs = ['Category1', 'Category2', 'Category4']

    You can also specify deeper levels to filter on, e.g.:
        include_subdirs = ['Category1/TopicA', 'Category1/TopicB']

    Args:
        dir_path (pathlib Path): Path object representing the directory
            to search.
        include_subdirs (list, optional): list of string paths to include
            in the filtered list of md files (e.g. ['p1', 'p2', 'p3/sp1']).
            If no list is specified, then no filtering is done on paths.
            Defaults to None.
        include_root (bool, optional): include files that are directly in
            the dir_path (root dir).  Defaults to True.

    Returns:
        list of Path objects
    """
    return get_relpaths_matching_subdirs(
        dir_path,
        extension='canvas',
        include_subdirs=include_subdirs,
        include_root=include_root)


def get_canvas_content(filepath: Path) -> dict:
    """Get JSON content from canvas file as a Python dict.

    Args:
        filepath (Path): Path object representing the canvas file.

    Returns:
        dict
    """
    with open(filepath, encoding='utf-8') as f:
        json_as_dict = json.load(f)
    return json_as_dict


def get_canvas_graph_detail(canvas_content: dict) -> \
        tuple[nx.MultiDiGraph,
              dict[str, tuple[int, int]],
              dict[tuple[str, str], str]]:
    """Get the content from a canvas in a NetworkX graph.  With all the
    detail that is returned, it is possible to recreate the layout of the
    material in the canvas.

    If you only care about the graph connections and you are not trying to
    visualise the canvas, use this pattern:
        G, _, _ = get_canvas_graph_detail(...)

    Args:
        canvas_content (dict): dict that represents the JSON content from a
            canvas file.

    Returns:
        G, pos, edge_labels:
            G: NetworkX graph
            pos: list of co-ordinates for each node in graph
            edge_labels: list of labels for each edge in graph
    """
    G = nx.MultiDiGraph()
    nodes_list = [i.get('id') for i in canvas_content['nodes']
                  if i.get('type') != 'group']
    G.add_nodes_from(nodes_list)

    graph_edges_list = [(i.get('fromNode'), i.get('toNode'))
                        for i in canvas_content['edges']
                        if i.get('type') != 'group']
    G.add_edges_from(graph_edges_list)

    # y co-ord needs to be flipped to reflect app(?):
    pos = {i.get('id'): (i.get('x'), -i.get('y'))
           for i in canvas_content['nodes']
           if i.get('type') != 'group'}

    edge_labels = dict([((i.get('fromNode'), i.get('toNode')),
                         f"{i.get('label', '')}")
                        for i in canvas_content['edges']])

    return G, pos, edge_labels



# WIKILINKS AND EMBEDDED FILES: regex that includes any aliases
# group 0 captures embedded link; group 1 is everything inside [[]]
WIKILINK_REGEX = r'(!)?\[{2}([^\]\]]+)\]{2}'

# TAGS
TAG_INCLUDE_NESTED_REGEX = r'(?<!\()(?<!\\)#{1}([A-z]+[0-9_\-]*[A-Z0-9]?[^\s]+(?![^\[\[]*\]\]))\/?'
TAG_MAIN_ONLY_REGEX = r'(?<!\()#{1}([A-z]+[0-9_\-]*[A-Z0-9]?)\/?'

# md links: catch URLs or paths
INLINE_LINK_AFTER_HTML_PROC_REGEX = r'\[([^\]]+)\]\(<([^)]+)>\)'
INLINE_LINK_VIA_MD_ONLY_REGEX = r'\[([^\]]+)\]\(([^)]+)\)'

# helpers:
WIKILINK_AS_STRING_REGEX = r'\[[^\]]+\]\([^)]+\)'
EMBEDDED_FILE_LINK_AS_STRING_REGEX = r'!?\[{2}([^\]\]]+)\]{2}'



import setuptools

with open("README.md", "r", encoding="utf8") as f:
    LONG_DESCRIPTION = f.read()

PROJECT_URLS = {"Source": "https://github.com/mfarragher/obsidiantools"}
INSTALL_REQUIRES = [
    "markdown",
    "pymdown-extensions",
    "html2text",
    "pandas",
    "numpy",
    "networkx",
    "python-frontmatter",
    "beautifulsoup4",
    "bleach",
    "lxml"]

CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Education",
    "Intended Audience :: Science/Research",
    "Topic :: Text Processing :: Markup :: Markdown",
    "Topic :: Office/Business",
    "Topic :: Education",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.9",
    "License :: OSI Approved :: BSD License",
    "Operating System :: OS Independent",
]

setuptools.setup(
    name="obsidiantools",
    version="0.9.0",
    author="Mark Farragher",
    description="Obsidian Tools - a Python interface for Obsidian.md vaults",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    keywords=["obsidian", "obsidian.md", "knowledge",
              "note-taking", "notes",
              "knowledge management",
              "connected notes"],
    url="https://github.com/mfarragher/obsidiantools",
    project_urls=PROJECT_URLS,
    packages=setuptools.find_packages(exclude=("tests")),
    python_requires=">=3.9",
    install_requires=INSTALL_REQUIRES,
    license="BSD",
    classifiers=CLASSIFIERS
)



import os
import networkx as nx
import numpy as np
import pandas as pd
from collections import Counter
from pathlib import Path

# init
from .md_utils import (get_md_relpaths_matching_subdirs)
from .canvas_utils import (get_canvas_relpaths_matching_subdirs)
# connect
from .md_utils import (_get_md_front_matter_and_content,
                       _get_html_from_md_content,
                       _get_md_links_from_source_text,
                       _get_unique_md_links_from_source_text,
                       _get_unique_wikilinks_from_source_text,
                       _get_all_wikilinks_from_source_text,
                       _get_all_embedded_files_from_source_text,
                       get_tags,
                       get_source_text_from_html,
                       _get_all_latex_from_html_content)
# gather
from .md_utils import (get_source_text_from_md_file,
                       get_readable_text_from_md_file)
# canvas:
from .canvas_utils import (get_canvas_content,
                           get_canvas_graph_detail)


class Vault:
    def __init__(self, dirpath: Path, *,
                 include_subdirs: list[str] = None, include_root: bool = True):
        """A Vault object lets you dig into your Obsidian vault, by giving
        you a toolkit for analysing its contents.  Specify a dirpath to
        instantiate the class.  This class is intended to support multiple
        operating systems so pass a pathlib Path object.

        By calling CONNECT you generate a graph of your vault.
        All the wikilinks in the vault are used to produce the graph, to
        reflect how analytics are displayed in Obsidian, where links are
        counted rather than notes.  For example, if a note has 2 backlinks
        then there will be 2 wikilinks (to another note) visible when you do
        analysis of the edges & nodes.
        Detail on canvas files is also stored when calling this method.

        By calling GATHER you store the plaintext of your vault in the notes
        index attribute.  You can specify rules for how the text is
        processed, e.g. whether code blocks should be removed.

        The API is fluent so you can set up your vault, by connecting
        and gathering your notes in one line:

        vault = Vault(dirpath).connect().gather()

        The class supports subdirectories and relies heavily on relative
        paths for the API.

        Args:
            dirpath (pathlib Path): the directory to analyse.  This would
                typically be the vault's directory.  If you have a
                subdirectory of the vault with notes you want to inspect,
                then you could pass that.
            include_subdirs (list, optional): list of string paths to include
                in the filtered list of md files (e.g. ['p1', 'p2', 'p3/sp1']).
                If no list is specified, then no filtering is done on paths.
                Defaults to None.
            include_root (bool, optional): include files that are directly in
                the dir_path (root dir).  Defaults to True.

        Methods for setup:
            connect: connect notes in a graph
            gather: gather text content of notes

        Methods for analysis of an individual note:
            get_backlinks
            get_backlink_counts
            get_wikilinks
            get_wikilink_counts
            get_embedded_files
            get_math
            get_front_matter
            get_md_links
            get_tags
            get_source_text
            get_readable_text

        Methods for analysis across multiple notes:
            get_note_metadata

        Attributes - general:
            dirpath (arg)
            is_connected
            is_gathered
        Attributes - md-related:
            file_index
            graph
            backlinks_index
            wikilinks_index
            unique_wikilinks_index
            embedded_files_index
            math_index
            md_links_index
            unique_md_links_index
            tags_index
            nonexistent_notes
            isolated_notes
            source_text_index
            readable_text_index
        Attributes - canvas-related:
            canvas_file_index
            canvas_content_index
            canvas_graph_detail_index
        """
        self._dirpath = dirpath
        self._file_index = self._get_md_relpaths_by_name(
            include_subdirs=include_subdirs,
            include_root=include_root)
        self._canvas_file_index = self._get_canvas_relpaths_by_name(
            include_subdirs=include_subdirs,
            include_root=include_root)

        self._is_connected = False
        self._is_gathered = False

        # via md content:
        self._graph = None
        self._backlinks_index = {}
        self._wikilinks_index = {}
        self._unique_wikilinks_index = {}
        self._embedded_files_index = {}
        self._math_index = {}
        self._md_links_index = {}
        self._unique_md_links_index = {}
        self._tags_index = {}
        self._nonexistent_notes = []
        self._isolated_notes = []
        self._front_matter_index = {}
        self._source_text_index = {}
        self._readable_text_index = {}

        # via canvas content:
        self._canvas_content_index = {}
        self._canvas_graph_detail_index = {}

    @property
    def dirpath(self) -> Path:
        """pathlib Path"""
        return self._dirpath

    @property
    def file_index(self) -> dict[str, Path]:
        """dict: one-to-one mapping of md filename (k) to relative path (v)"""
        return self._file_index

    @file_index.setter
    def file_index(self, value) -> dict[str, Path]:
        self._file_index = value

    @property
    def canvas_file_index(self) -> dict[str, Path]:
        """dict: one-to-one mapping of canvas filename (k) to relative path
        (v)"""
        return self._canvas_file_index

    @canvas_file_index.setter
    def canvas_file_index(self, value) -> dict[str, Path]:
        self._canvas_file_index = value

    @property
    def graph(self) -> nx.MultiDiGraph:
        """networkx Graph"""
        return self._graph

    @property
    def backlinks_index(self) -> dict[str, list[str]]:
        """dict of lists: note name (k) to lists (v).  v is [] if k
        has no backlinks."""
        return self._backlinks_index

    @property
    def wikilinks_index(self) -> dict[str, list[str]]:
        """dict of lists: filename (k) to lists (v).  v is [] if k
        has no wikilinks."""
        return self._wikilinks_index

    @property
    def unique_wikilinks_index(self) -> dict[str, list[str]]:
        """dict of lists: filename (k) to lists (v).  v is [] if k
        has no wikilinks."""
        return self._unique_wikilinks_index

    @property
    def embedded_files_index(self) -> dict[str, list[str]]:
        """dict: note name (k) to list of embedded file string (v).
        v is [] if k has no embedded files."""
        return self._embedded_files_index

    @property
    def math_index(self) -> dict[str, list[str]]:
        """dict: note name (k) to list of LaTeX math string (v).  v is [] if
        k has no LaTeX."""
        return self._math_index

    @property
    def md_links_index(self) -> dict[str, list[str]]:
        """dict of lists: filename (k) to lists (v).  v is [] if k
        has no markdown links."""
        return self._md_links_index

    @property
    def unique_md_links_index(self) -> dict[str, list[str]]:
        """dict of lists: filename (k) to lists (v).  v is [] if k
        has no markdown links."""
        return self._unique_md_links_index

    @property
    def tags_index(self) -> dict[str, list[str]]:
        """dict of lists: filename (k) to lists (v).  v is [] if k
        has no tags."""
        return self._tags_index

    @property
    def nonexistent_notes(self) -> list[str]:
        """list: notes without files, i.e. the notes have backlink(s) but
        their md files don't exist yet.

        They are ideas floating around in the Obsidian graph... waiting to
        be created as actual notes one day :-)"""
        return self._nonexistent_notes

    @property
    def isolated_notes(self) -> list[str]:
        """list: notes (with their own md files) that lack backlinks and
        lack wikilinks.  They are not connected to other notes in the
        Obsidian graph at all."""
        return self._isolated_notes

    @property
    def front_matter_index(self) -> dict[str, list[str]]:
        """dict: note name (k) to front matter (v).  v is {} if no front
        matter was extracted from note."""
        return self._front_matter_index

    @property
    def is_connected(self) -> bool:
        """Bool: has the connect function been called to set up graph?"""
        return self._is_connected

    @property
    def is_gathered(self) -> bool:
        """Bool: has the gather function been called to gather text?"""
        return self._is_gathered

    @property
    def source_text_index(self) -> dict[str, str]:
        """dict of strings: filename (k) to source text string (v).  v is ''
        if k has no text.

        Source text aims to preserve as much of the text written to a file
        as possible.  This is done through the use of pymarkdown extensions
        to maintain the additional markdown features that Obsidian uses."""
        return self._source_text_index

    @source_text_index.setter
    def source_text_index(self, value) -> dict[str, str]:
        self._source_text_index = value

    @property
    def readable_text_index(self) -> dict[str, str]:
        """dict of strings: filename (k) to 'readable' string (v).  v is ''
        if k has no text.

        Readable text has all the main formatting removed.  This makes it
        easier to use the text for NLP analysis, with minimal processing
        needed, while still maintaining the meaning of text."""
        return self._readable_text_index

    @readable_text_index.setter
    def readable_text_index(self, value) -> dict[str, str]:
        self._readable_text_index = value

    @property
    def canvas_content_index(self) -> dict[str, str]:
        """dict of dict: 'shortest path when possible' filepath with canvas
        ext (k), to canvas content dict (v).  v is {} if k has no content."""
        return self._canvas_content_index

    @canvas_content_index.setter
    def canvas_content_index(self, value) -> dict[str, str]:
        self._canvas_content_index = value

    @property
    def canvas_graph_detail_index(self) -> \
        dict[str, tuple[nx.MultiDiGraph,
                        dict[str, tuple[int, int]],
                        dict[tuple[str, str], str]]
             ]:
        """dict of tuple: 'shortest path when possible' filepath with canvas
        ext (k), to canvas graph detail tuple (v)."""
        return self._canvas_graph_detail_index

    @canvas_graph_detail_index.setter
    def canvas_graph_detail_index(self, value) -> \
        dict[str, tuple[nx.MultiDiGraph,
                        dict[str, tuple[int, int]],
                        dict[tuple[str, str], str]]
             ]:
        self._canvas_graph_detail_index = value

    def connect(self, *, show_nested_tags: bool = False):
        """connect your notes together by representing the vault as a
        Networkx graph object, G.

        With your vault instantiated, set up the graph through:
            vault.connect()

        The graph G is written to the 'graph' attribute, to represent the
        Obsidian graph of your notes.

        Args:
            show_nested_tags (Boolean): show nested tags in the output.
                Defaults to False (which would mean only the highest level
                of any nested tags are included in the output).
        """
        # md content:
        if not self._is_connected:
            # index dicts, where k is a note name in the vault:
            md_links_ix = {}
            md_links_unique_ix = {}
            embedded_files_ix = {}
            tags_ix = {}
            math_ix = {}
            front_matter_ix = {}
            wikilinks_ix = {}
            wikilinks_unique_ix = {}

            # loop through md files:
            for f, relpath in self._file_index.items():
                # MAIN file read:
                front_matter, content = _get_md_front_matter_and_content(
                    self._dirpath / relpath)
                html = _get_html_from_md_content(content)
                src_txt = get_source_text_from_html(
                    html, remove_code=True)

                # info from core text:
                md_links_ix[f] = _get_md_links_from_source_text(src_txt)
                md_links_unique_ix[f] = _get_unique_md_links_from_source_text(src_txt)
                embedded_files_ix[f] = _get_all_embedded_files_from_source_text(
                    src_txt, remove_aliases=True)
                wikilinks_ix[f] = _get_all_wikilinks_from_source_text(
                    src_txt, remove_aliases=True)
                wikilinks_unique_ix[f] = _get_unique_wikilinks_from_source_text(
                    src_txt, remove_aliases=True)
                # info from html:
                math_ix[f] = _get_all_latex_from_html_content(html)
                # split out front matter:
                front_matter_ix[f] = front_matter

                # MORE file reads needed for extra info:
                tags_ix[f] = get_tags(self._dirpath / relpath,
                                      show_nested=show_nested_tags)

            self._md_links_index = md_links_ix
            self._unique_md_links_index = md_links_unique_ix
            self._embedded_files_index = embedded_files_ix
            self._tags_index = tags_ix
            self._math_index = math_ix
            self._front_matter_index = front_matter_ix
            # to be used for graph:
            self._wikilinks_index = wikilinks_ix
            self._unique_wikilinks_index = wikilinks_unique_ix

            # graph:
            G = nx.MultiDiGraph(wikilinks_ix)
            self._graph = G
            # info obtained from graph:
            self._backlinks_index = self._get_backlinks_index(graph=G)
            self._nonexistent_notes = self._get_nonexistent_notes()
            self._isolated_notes = self._get_isolated_notes(graph=G)

            self._is_connected = True

        # canvas content:
        # loop through canvas files:
        canvas_content_ix = {}
        canvas_graph_detail_ix = {}
        for f, relpath in self._canvas_file_index.items():
            content_c = get_canvas_content(
                self._dirpath / relpath)
            canvas_content_ix[f] = content_c
            G_c, pos_c, edge_labels_c = get_canvas_graph_detail(
                content_c)
            canvas_graph_detail_ix[f] = G_c, pos_c, edge_labels_c
        self._canvas_content_index = canvas_content_ix
        self._canvas_graph_detail_index = canvas_graph_detail_ix

        return self  # fluent

    def gather(self, *, tags: list[str] = None):
        """gather the content of your notes so that all the plaintext is
        stored in one place for easy access.

        The content of each note is stored in the source_text_index attribute.
        This enables text from the files to be accessible in one place.

        The stripped down content of the notes' text is stored in the
        readable_text_index attribute.  This enables text to be used more
        easily in NLP analysis.

        With your vault connected, gather your note content through:
            vault.gather()

        Args:
            tags (list, optional): Defaults to None, to keep all headers
                and paragraph formatting in the readable_text for notes.
                Otherwise specify a list of HTML tags to use to preserve
                their formatting in the final text.  For example, tags=[]
                will remove all header formatting (e.g. '#', '##' chars)
                and produces a one-line string.
        """
        # source text will not remove any content:
        self._source_text_index = {
            k: get_source_text_from_md_file(self._dirpath / v,
                                            remove_code=True,
                                            remove_math=True)
            for k, v in self._file_index.items()}
        self._readable_text_index = {
            k: get_readable_text_from_md_file(self._dirpath / v,
                                              tags=tags)
            for k, v in self._file_index.items()}

        self._is_gathered = True

        return self  # fluent

    def get_backlinks(self, note_name: str) -> list[str]:
        """Get backlinks for a note (given its name).

        If a note has not been created, but has wikilinks pointing to it
        elsewhere in the vault, then it will return those backlinks.

        If a note is not in the graph at all, then the function will raise
        an AttributeError.

        Args:
            note_name (str): the string that is the name in the graph.
                This is NOT a filepath!

        Returns:
            list
        """
        if not self._is_connected:
            raise AttributeError('Connect notes before calling the function')

        if note_name not in self._graph.nodes:
            raise ValueError('"{}" not found in graph.'.format(note_name))
        else:
            return self._backlinks_index[note_name]

    def get_backlink_counts(self, note_name: str) -> dict[str, int]:
        """Get counts of backlinks for a note (given its name).

        Args:
            note_name (str): the string that is the name in the graph.
                This is NOT a filepath!

        Returns:
            dict of integers >= 1
        """
        if not self._is_connected:
            raise AttributeError('Connect notes before calling the function')

        if note_name not in self._graph.nodes:
            raise ValueError('"{}" not found in graph.'.format(note_name))
        else:
            backlinks = self.get_backlinks(note_name)
            return dict(Counter(backlinks))

    def get_wikilinks(self, file_name: str) -> list[str]:
        """Get wikilinks for a note (given its filename).

        Wikilinks can only appear in notes that already exist, so if a
        note is not in the file_index at all then the function will raise
        a ValueError.

        Args:
            file_name (str): the filename string that is in the file_index.
                This is NOT the filepath!

        Returns:
            list
        """
        if not self._is_connected:
            raise AttributeError('Connect notes before calling the function')

        if file_name not in self._file_index:
            raise ValueError('"{}" does not exist so it cannot have wikilinks.'.format(file_name))
        else:
            return self._wikilinks_index[file_name]

    def get_wikilink_counts(self, note_name: str) -> dict[str, int]:
        """Get counts of wikilinks for a note (given its name).

        Args:
            note_name (str): the string that is the name in the graph.
                This is NOT a filepath!

        Returns:
            dict of integers >= 1
        """
        if not self._is_connected:
            raise AttributeError('Connect notes before calling the function')

        if note_name not in self._graph.nodes:
            raise ValueError('"{}" not found in graph.'.format(note_name))
        else:
            wikilinks = self.get_wikilinks(note_name)
            return dict(Counter(wikilinks))

    def get_embedded_files(self, file_name: str) -> list[str]:
        """Get embedded files for a note (given its filename).

        Embedded files can only appear in notes that already exist, so if a
        note is not in the file_index at all then the function will raise
        a ValueError.

        Args:
            file_name (str): the filename string that is in the file_index.
                This is NOT the filepath!

        Returns:
            list
        """
        if not self._is_connected:
            raise AttributeError('Connect notes before calling the function')

        if file_name not in self._file_index:
            raise ValueError('"{}" does not exist so it cannot have embedded files.'.format(file_name))
        else:
            return self._embedded_files_index[file_name]

    def get_md_links(self, file_name: str) -> list[str]:
        """Get markdown links for a note (given its filename).

        Markdown links can only appear in notes that already exist, so if a
        note is not in the file_index at all then the function will raise
        a ValueError.

        Args:
            file_name (str): the filename string that is in the file_index.
                This is NOT the filepath!

        Returns:
            list
        """
        if not self._is_connected:
            raise AttributeError('Connect notes before calling the function')

        if file_name not in self._file_index:
            raise ValueError('"{}" does not exist so it cannot have md links.'.format(file_name))
        else:
            return self._md_links_index[file_name]

    def get_front_matter(self, file_name: str) -> list[dict]:
        """Get front matter for a note (given its filename).

        Front matter can only appear in notes that already exist, so if a
        note is not in the file_index at all then the function will raise
        a ValueError.

        Args:
            file_name (str): the filename string that is in the file_index.
                This is NOT the filepath!

        Returns:
            list
        """
        if not self._is_connected:
            raise AttributeError('Connect notes before calling the function')
        if file_name not in self._file_index:
            raise ValueError('"{}" does not exist so it cannot have front matter.'.format(file_name))
        else:
            return self._front_matter_index[file_name]

    def get_tags(self, file_name: str, *,
                 show_nested: bool = False) -> list[str]:
        """Get tags for a note (given its filename).
        By default, only the highest level of any nested tags is shown
        in the output.

        Tags can only appear in notes that already exist, so if a
        note is not in the file_index at all then the function will raise
        a ValueError.

        Args:
            file_name (str): the filename string that is in the file_index.
                This is NOT the filepath!

        Returns:
            list
        """
        if not self._is_connected:
            raise AttributeError('Connect notes before calling the function')
        if file_name not in self._file_index:
            raise ValueError('"{}" does not exist so it cannot have tags.'.format(file_name))
        else:
            return self._tags_index[file_name]

    def get_source_text(self, file_name: str) -> str:
        """Get text for a note (given its filename).  This requires the vault
        function 'gather' to have been called.

        Text can only appear in notes that already exist, so if a
        note is not in the file_index at all then the function will raise
        a ValueError.

        Args:
            file_name (str): the filename string that is in the file_index.
                This is NOT the filepath!

        Returns:
            str
        """
        if not self._is_gathered:
            raise AttributeError('Gather notes before calling the function')
        if file_name not in self._file_index:
            raise ValueError('"{}" does not exist so it cannot have text.'.format(file_name))
        else:
            return self._source_text_index[file_name]

    def get_readable_text(self, file_name: str) -> str:
        """Get readable text for a note (given its filename).
        This requires the vault function 'gather' to have been called.

        Change the arguments of the 'gather' function to specify how the text
        output should be stored, e.g. whether all HTML tags should be removed
        before the final output.

        Text can only appear in notes that already exist, so if a
        note is not in the file_index at all then the function will raise
        a ValueError.

        Args:
            file_name (str): the filename string that is in the file_index.
                This is NOT the filepath!

        Returns:
            str
        """
        if not self._is_gathered:
            raise AttributeError('Gather notes before calling the function')
        if file_name not in self._file_index:
            raise ValueError('"{}" does not exist so it cannot have text.'.format(file_name))
        else:
            return self._readable_text_index[file_name]

    def _get_md_relpaths(self, **kwargs) -> list[Path]:
        """Return list of md filepaths *relative* to the directory instantiated
        for the class.

        Returns:
            list
        """
        return get_md_relpaths_matching_subdirs(self._dirpath, **kwargs)

    def _get_canvas_relpaths(self, **kwargs) -> list[Path]:
        """Return list of canvas filepaths *relative* to the directory instantiated
        for the class.

        Returns:
            list
        """
        return get_canvas_relpaths_matching_subdirs(self._dirpath,
                                                    **kwargs)

    def __get_relpaths_by_name(self, *, extension, **kwargs) -> dict[str, Path]:
        """Return k,v pairs
        where k is the file name
        and v is the relpath of the {extension} file

        Returns:
            dict
        """
        # get note names (for k) and relpaths (for v):
        # keep or remove ext, based on how the file refs appear in wikilinks:
        if extension == 'md':
            relpaths_list = self._get_md_relpaths(
                **kwargs)
            # remove .md ext:
            all_file_names_list = [f.stem for f in relpaths_list]
        if extension == 'canvas':
            relpaths_list = self._get_canvas_relpaths(
                **kwargs)
            # keep .canvas ext:
            all_file_names_list = [f.name for f in relpaths_list]

        # get indices of dupe note names:
        _, inverse_ix, counts = np.unique(
            np.array(all_file_names_list),
            return_inverse=True,
            return_counts=True,
            axis=0)
        dupe_names_ix = np.where(counts[inverse_ix] > 1)[0]

        # get shortest paths via mask:
        shortest_paths_arr = np.array(all_file_names_list, dtype=object)
        if extension == 'md':
            shortest_paths_arr[dupe_names_ix] = np.array(
                [str(fpath.with_suffix(''))
                 for fpath in relpaths_list])[dupe_names_ix]
        if extension == 'canvas':
            shortest_paths_arr[dupe_names_ix] = np.array(
                [str(fpath)
                 for fpath in relpaths_list])[dupe_names_ix]
        return {n: p for n, p in zip(shortest_paths_arr, relpaths_list)}

    def _get_md_relpaths_by_name(self, **kwargs) -> dict[str, Path]:
        return self.__get_relpaths_by_name(extension='md',
                                           **kwargs)

    def _get_canvas_relpaths_by_name(self, **kwargs) -> dict[str, Path]:
        return self.__get_relpaths_by_name(extension='canvas',
                                           **kwargs)

    def _get_backlinks_index(self, *,
                             graph: nx.MultiDiGraph) -> dict[str, list[str]]:
        """Return k,v pairs
        where k is the md note name
        and v is list of ALL backlinks found in k"""
        return {n: [n[0] for n in list(graph.in_edges(n))]
                for n in self._graph.nodes}

    def get_note_metadata(self) -> pd.DataFrame:
        """Structured dataset of metadata on the vault's notes.  This
        includes filepaths and counts of different link types.

        The df is indexed by 'note' (i.e. nodes in the graph).

        Notes that haven't been created will only have info on the number
        of backlinks - other columns will have NaN.

        Returns:
            pd.DataFrame
        """

        if not self._is_connected:
            raise AttributeError('Connect notes before calling the function')

        df = (pd.DataFrame(index=list(self._backlinks_index.keys()))
              .rename_axis('note')
              .pipe(self._create_note_metadata_columns)
              .pipe(self._clean_up_note_metadata_dtypes)
              )
        return df

    def _create_note_metadata_columns(self,
                                      df: pd.DataFrame) -> pd.DataFrame:
        """pipe func for mutating df"""
        df['rel_filepath'] = [self._file_index.get(f, np.NaN)
                              for f in df.index]
        df['abs_filepath'] = np.where(df['rel_filepath'].notna(),
                                      [self._dirpath / Path(str(f))
                                       for f in df['rel_filepath']],
                                      np.NaN)
        df['note_exists'] = np.where(df['rel_filepath'].notna(),
                                     True, False)
        df['n_backlinks'] = [len(self.get_backlinks(f)) for f in df.index]
        df['n_wikilinks'] = np.where(df['note_exists'],
                                     [len(self._wikilinks_index.get(f, []))
                                     for f in df.index],
                                     np.NaN)
        df['n_tags'] = np.where(df['note_exists'],
                                [len(self._tags_index.get(f, []))
                                 for f in df.index],
                                np.NaN)
        df['n_embedded_files'] = np.where(df['note_exists'],
                                          [len(self._embedded_files_index.get(
                                              f, []))
                                           for f in df.index],
                                          np.NaN)
        df['modified_time'] = pd.to_datetime(
            [os.path.getmtime(f) if not pd.isna(f) else np.NaN
             for f in df['abs_filepath']],
            unit='s'
        )
        return df

    def _clean_up_note_metadata_dtypes(self,
                                       df: pd.DataFrame) -> pd.DataFrame:
        """pipe func for mutating df"""
        df['rel_filepath'] = np.where(df['rel_filepath'].notna(),
                                      [Path(str(f))
                                       for f in df['rel_filepath']],
                                      np.NaN)
        df['n_wikilinks'] = df['n_wikilinks'].astype(float)  # for consistency
        return df

    def _get_nonexistent_notes(self) -> list[str]:
        """Get notes that have backlinks but don't have md files.

        The comparison is done with sets but the result is returned
        as a list."""
        return list(set(self.backlinks_index.keys())
                    .difference(set(self.file_index)))

    def _get_isolated_notes(self, *,
                            graph: nx.MultiDiGraph) -> list[str]:
        """Get notes that are not connected to any other notes in the vault,
        i.e. they have 0 wikilinks and 0 backlinks.

        These notes are retrieved from the graph."""
        return list(nx.isolates(graph))



from pathlib import Path
from glob import glob


def get_relpaths_from_dir(dir_path: Path, *, extension: str) -> list[Path]:
    """Get list of relative paths for {extension} files in a given directory,
    including any subdirectories.

    Thus if the vault directory is the argument, then the function returns
    a list of all the {extension} files found in the vault.

    Args:
        dir_path (pathlib Path): Path object representing the directory
            to search.
        extension (str): file extension like 'md' or 'canvas'.

    Returns:
        list of Path objects
    """
    relpaths_list = [Path(p).relative_to(dir_path)
                     for p in glob(f"{dir_path}/**/*.{extension}",
                     recursive=True)]
    return relpaths_list


def get_relpaths_matching_subdirs(dir_path: Path, *,
                                  extension: str,
                                  include_subdirs: list = None,
                                  include_root: bool = True) -> list[Path]:
    """Get list of relative paths for {extension} files in a given directory,
    filtered to include specified subdirectories (with include_subdirs
    kwarg).  The default arguments align with get_relpaths_from_dir
    function, but this function enables more flexibility.

    For example, if you had a vault with folders named by category, and
    filter them like this in Obsidian:
        path:Category1/ OR path:Category2/ OR path:Category4/
    then you can use the include_subdirs kwarg to do that with this function:
        include_subdirs = ['Category1', 'Category2', 'Category4']

    You can also specify deeper levels to filter on, e.g.:
        include_subdirs = ['Category1/TopicA', 'Category1/TopicB']

    Args:
        dir_path (pathlib Path): Path object representing the directory
            to search.
        extension (str): file extension like 'md' or 'canvas'.
        include_subdirs (list, optional): list of string paths to include
            in the filtered list of md files (e.g. ['p1', 'p2', 'p3/sp1']).
            If no list is specified, then no filtering is done on paths.
            Defaults to None.
        include_root (bool, optional): include files that are directly in
            the dir_path (root dir).  Defaults to True.

    Returns:
        list of Path objects
    """
    # Obsidian's 'shortest path' for files uses forward slash across
    # operating systems, so as_posix() is used to yield paths with
    # forward slash consistently here.

    if include_subdirs:
        include_subdirs_final = [str(Path(i).as_posix())
                                 for i in include_subdirs]

    if not include_subdirs and include_root:
        return get_relpaths_from_dir(dir_path,
                                     extension=extension)
    elif not include_subdirs and not include_root:
        return [i for i in get_relpaths_from_dir(dir_path,
                                                 extension=extension)
                if str(i.parent.as_posix()) != '.']
    else:
        if include_root:
            return [i for i in get_relpaths_from_dir(dir_path,
                                                     extension=extension)
                    if str(i.parent.as_posix())
                    in include_subdirs_final + ['.']]
        else:
            return [i for i in get_relpaths_from_dir(dir_path,
                                                     extension=extension)
                    if str(i.parent.as_posix())
                    in include_subdirs_final]
